#include "PointCloudSurface.h"
#include "PointSearch.h"
#include "Grid.h"
#include "utils.h"


using namespace std;




PointCloudSurface::PointCloudSurface(Grid& input): scalars(input.numPoints()), normals(input.numPoints())
{
   int NP = input.numPoints();
   	
   points.resize(NP);
   
   for(int i=0;i<NP;++i)						//1. Copy all points from the input grid into this
   {											//   Modify this to initialize a point cloud differently, e.g.
	   Point3d p;								//   by explicitly passing on the points from another data source
	   input.getPoint(i,p.data);
	   points[i] = p;
   }	   

   const int numNeighbors = 20;					//2. Initialize a spatial search engine on the cloud points. We'll use
												//   this next to find point nearest-neighbors to estimate normals,	
   point_search = new PointSearch(input,numNeighbors);  //and also to implement findCell()

   float norm[3], *nrm;
   Point3d u1,u2;
   vector<Point3d> nbs; nbs.reserve(numNeighbors);	  

   for(int pid=0;pid<NP;++pid)					//3. Estimate surface normals for the surface points
   {
	  Point3d& p = points[pid];	  
	  nbs.clear();								

	  vector<int>   indexes;
	  vector<float> distances;					//3.1. Get closest numNeighbors neighbors of each cloud point
	  point_search->closest(numNeighbors,p,indexes,distances);

	  for(int i=0;i<numNeighbors;++i)
		nbs.push_back(points[indexes[i]]);

												//3.2. Estimate surface normal by PCA on each point's neighborhood
												//     The normal is the smallest eigenvector

	  if (nbs.size()>=3)						//3.3. Estimate plane only if there are enough neighbors...
	  {
	    float cl = plane(nbs,nrm=norm,u1.data,u2.data,0);
	    if (cl<0.2) nrm = 0;					//     Anisotropy is planar: use plane normal, else fall back on default-normal
	  }
	  else nrm = 0;

	  Point3d n = (nrm)? Point3d(nrm) : Point3d();
      normals.setC0Vector(pid,n.data);	

	  u1.normalize();							//3.4. Estimate a good radius for the neighborhood. For this, project all points
												//     along the longest eigenvector u1, and compute the maximal projection-length	
	  float max_dist = 0;
	  for(int i=0;i<numNeighbors;++i)
	  {
	    Point3d dir = nbs[i]-p;
		max_dist = max(max_dist,fabs(dir.dot(u1)));
	  }		
	  
	  p.value = max_dist/2; 	  
   }
}


PointCloudSurface::~PointCloudSurface()
{
	delete point_search;
}
	

int	PointCloudSurface::findCell(float* p)
{
	  vector<int>   indexes;
	  vector<float> distances;					//3.1. Get closest numNeighbors neighbors of each cloud point

	  point_search->closest(1,Point3d(p),indexes,distances);
	  
	  return indexes[0];
}
