#include "PointCloudSurfaceRenderer.h"
#include "PointCloudSurface.h"
#include "utils.h"
#include "zpr.h"
#include <GLUT/glut.h>
#include <math.h>
#include <string.h>
#include <string>
#include <iostream>

using namespace std;



static const float def_color[] = {0.5,0.5,0.5};						//Default surface color when not using color-mapping



static char* textFileRead(char* buffer,const string& fn)			//
{
  FILE* fp = fopen(fn.c_str(),"rt");
  if (!fp) return 0;
   
  fseek(fp, 0, SEEK_END);
  int count = ftell(fp);
  rewind(fp);
  if (count<=0) return 0; 
      
  count = fread(buffer,sizeof(char),count,fp);
  buffer[count] = '\0';   
  fclose(fp);
 
  return buffer;
}



static void printShaderLog(GLuint obj) 
{
  GLint infoLogLength = 0;
  GLsizei charsWritten  = 0;
  GLchar *infoLog;

  glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infoLogLength);

  if (infoLogLength > 1)
  {
    infoLog = (char *) malloc(infoLogLength);
    glGetShaderInfoLog(obj, infoLogLength, &charsWritten, infoLog);
    printf("\n%s\n",infoLog);
    free(infoLog);
  }
}



static void printProgramLog(GLuint obj) 
{
  GLint infoLogLength = 0;
  GLsizei charsWritten  = 0;
  GLchar *infoLog;
  
  glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infoLogLength);
  
  if (infoLogLength > 1)
  {
    infoLog = (char *) malloc(infoLogLength);
    glGetProgramInfoLog(obj, infoLogLength, &charsWritten, infoLog);
    printf("\n%s\n",infoLog);
    free(infoLog);
  }
}





PointCloudSurfaceRenderer::PointCloudSurfaceRenderer()
{
   sphere_faces   = 8;											//resolution of the GLUT spheres (small=fast)
   light_specular = 0.7;										//specular amount
   light_diffuse  = 0.7;										//diffuse amount
   specular_power = 10;											//shininess
   use_mipmaps    = true;										//use texture mipmaps for splatting or not
   draw_mode      = DRAW_POINTS;								//draw surface as simple points
   use_floating_textures = true;								//whether to use floating-point or 24-bit fixed-point textures for the splat height
   light_pos[0]	  = 0;											//light direction
   light_pos[1]   = 0;											//light position and type
   light_pos[2]   = 1;
   light_pos[3]   = 1;	
   ball_dt_offset = 0.0;										//correction factor for the size of the balls
   splat_dt_offset = 0.5;										//correction factor for the size of the splats
   splat_size = 512;											//size of largest texture splat; 512 seems quite enough
   ball_size = 2.0;												//ball radius for constant-radius surface drawing. Bit larger than a voxel's size
   use_colormap = false;										//whether to use scalar color-mapping on the surface or not
   splat_radius = 5;											//radius (pixels) for the splat used by drawSplats() 

   if(!(GL_ARB_fragment_program && GL_ARB_vertex_program && GL_EXT_geometry_shader4))
   {
     cerr<<"Error: Unable to load GL extensions:"<<endl;
	 cerr<<" GL_ARB_fragment_program, GL_ARB_vertex_program, GL_EXT_geometry_shader4"<<endl;
	 cerr<<"Exiting."<<endl;
     exit(-1);
   }      
   
   glGenTextures(2,balltex);									//Generate a texture-ID for the ball splat

   char buffer[10000];
   glGenProgramsARB(3, ball_shader);										//Create 5 shaders:
         
   ball_shader[0] = glCreateShader(GL_VERTEX_SHADER);						//3. Vertex shader, sprite-based splatting
   char* vs = textFileRead(buffer,"Shaders/ball_shader.vert");
   glShaderSource(ball_shader[0], 1,(const char**)&vs, 0);
   glCompileShader(ball_shader[0]);
   printShaderLog(ball_shader[0]);
                     
   ball_shader[1] = glCreateShader(GL_FRAGMENT_SHADER);						//4. Fragment shader, sprite-based splatting
   char *fs = textFileRead(buffer,"Shaders/ball_shader.frag");
   glShaderSource(ball_shader[1], 1, (const char**)&fs, 0);
   glCompileShader(ball_shader[1]);
   printShaderLog(ball_shader[1]);
            
   ball_shader[2] = glCreateShader(GL_GEOMETRY_SHADER_EXT);					//5. Geometry shader, sprite-based splatting
   char *gs = textFileRead(buffer,"Shaders/ball_shader.geom");
   glShaderSource(ball_shader[2], 1, (const char**)&gs, 0);
   glCompileShader(ball_shader[2]);
   printShaderLog(ball_shader[2]); 
   
   ball_prog = glCreateProgram();											//Assemble last 3 shaders in a 'program'
   glAttachShader(ball_prog,ball_shader[0]);
   glAttachShader(ball_prog,ball_shader[1]);
   glAttachShader(ball_prog,ball_shader[2]);
   glProgramParameteriEXT(ball_prog,GL_GEOMETRY_INPUT_TYPE_EXT,GL_POINTS);
   glProgramParameteriEXT(ball_prog,GL_GEOMETRY_OUTPUT_TYPE_EXT,GL_TRIANGLE_STRIP);
   int temp = 4;
   glProgramParameteriEXT(ball_prog,GL_GEOMETRY_VERTICES_OUT_EXT,temp);
   glLinkProgram(ball_prog);
   
   makeTextures();
   
   initGL();
}


void PointCloudSurfaceRenderer::initGL()
{      
  float lDiff[4], lSpec[4], lAmb[4];
  lAmb[0]  = lAmb[1]  = lAmb[2]  = 0; lAmb[3] = 0;
  lDiff[0] = lDiff[1] = lDiff[2] = light_diffuse; lDiff[3]=1;
  lSpec[0] = lSpec[1] = lSpec[2] = lSpec[3] = light_specular;
  
  glEnable(GL_COLOR_MATERIAL);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, lAmb);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, lDiff);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, lSpec);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, specular_power);

  GLfloat white[] = {1,1,1,1};	
  glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, white);
  glLightfv(GL_LIGHT0, GL_SPECULAR, white);
  glEnable(GL_NORMALIZE);
  glEnable(GL_AUTO_NORMAL);
  glEnable(GL_DEPTH_TEST);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
}



void PointCloudSurfaceRenderer::createSplatFloat(int SZ,GLfloat* img,int level)  //Creates a single SZ*SZ RGBA texture splat (floating-point)
{
   const float half = 0.5;
   const float C = SZ/2.0;
   const float C2 = C*C;										//Generate a half-sphere height profile, encode it as a luminance-alpha texture
   for(int i=0,idx=0;i<SZ;++i)									//The height should be normalized in [0..1]; alpha indicates the valid vs nonvalid pixels
     for(int j=0;j<SZ;++j,idx+=4)
	 {
	   float x  = i-C+half, y = j-C+half;
	   int   r2 = x*x+y*y;
	   if (r2>C2)												//Point outside the ball: color irrelevant, and it's transparent	
	   { 
	      img[idx] = img[idx+1] = img[idx+2] = img[idx+3] = 0; 
	   }
	   else														//Point in the ball: store a RGBA texel with:
	   {														// - R:  ball height [0..1] 
	          float   D = sqrt(C2-r2);							// - G:  ball shading [0..1]
		  float dfx = -x/D, dfy = -y/D;							// - A:  transparency mask (1)
		  Point3d n(dfx,dfy,1);									//  -B:  ball height, high-precision part [0..1/255]
		  n.normalize();										//This is the ball's normal vector
		  
		  float dp   = fabs(n.dot(light_pos));					//Simple Phong lighting model using same params as the OpenGL pipeline
		  float I    = light_diffuse*dp + light_specular*pow(dp,specular_power);
		  if (I>1) I=1;
	          img[idx]   = D/C;									//floating-point height	
		  img[idx+1] = 0;  
		  img[idx+2] = 0;  
		  img[idx+3] = I;										//intensity
	   }
	 }
   
   glTexImage2D(GL_TEXTURE_2D, level, GL_LUMINANCE_ALPHA32F_ARB, SZ, SZ, 0, GL_RGBA, GL_FLOAT, img);
   //We use a luminance-alpha internal format, since it's apparently cheaper and faster than RGBA
}





void PointCloudSurfaceRenderer::createSplatRGBA(int SZ,GLubyte* img,int level)		//Creates a single SZ*SZ RGBA texture splat (8-bit per channel)
{
	const float half = 0.5;
	const float C = SZ/2.0;
	const float C2 = C*C;											//Generate a half-sphere height profile, encode it as a luminance-alpha texture
	for(int i=0,idx=0;i<SZ;++i)										//The height should be normalized in [0..1]; alpha indicates the valid vs nonvalid pixels
		for(int j=0;j<SZ;++j,idx+=4)
		{
			float x  = i-C+half, y = j-C+half;
			int   r2 = x*x+y*y;
			if (r2>C2)												//Point outside the ball: color irrelevant, and it's transparent	
			{ 
				img[idx] = img[idx+1] = img[idx+2] = img[idx+3] = 0; 
			}
			else													//Point in the ball: store a RGBA texel with:
			{														// - R:  ball height [0..1] 
				float   D = sqrt(C2-r2);							// - G:  ball shading [0..1]
				float dfx = -x/D, dfy = -y/D;						// - A:  transparency mask (1)
				Point3d n(dfx,dfy,1);								//  -B:  ball height, high-precision part [0..1/255]
				n.normalize();										//This is the ball's normal vector
				
				float dp   = fabs(n.dot(light_pos));				//Simple Phong lighting model using same params as the OpenGL pipeline
				float I    = light_diffuse*dp + light_specular*pow(dp,specular_power);	
				if (I>1) I=1;
				
				unsigned int v = 255.0*255*255*(D/C);				//24-bit fixed-point height, normalized on [0..255^3-1]
				
				img[idx+2] = v % 256; v >>= 8;						//bits 23-16 of fixed-point height
				img[idx+1] = v % 256; v >>= 8;						//bits 15-8 of fixed-point height
				img[idx  ] = v % 256;								//bits 7-0 of fixed-point height
				img[idx+3] = 255*I;									//intensity
			}
		}


	glTexImage2D(GL_TEXTURE_2D, level, GL_RGBA, SZ, SZ, 0, GL_RGBA, GL_UNSIGNED_BYTE, img);
}


int PointCloudSurfaceRenderer::getScalingMode() const
{
  if (scaling_mode==-1) return (use_mipmaps)? GL_LINEAR_MIPMAP_LINEAR:GL_LINEAR;
  return scaling_mode;	
}


void PointCloudSurfaceRenderer::makeTextures(int sm)					//Generate splatting textures 
{
   scaling_mode = sm;											//Remember if we specified a scaling mode or using 'defaults'

   GLfloat* img = new GLfloat[4*splat_size*splat_size];			//Allocate some buffer where we'll synthesize the texture images
   GLubyte* imgb = (GLubyte*)img;  

   glBindTexture(GL_TEXTURE_2D, balltex[0]);					//Create mipmap levels by hand (to be accurate) for both the fixed-point
   for(int l=0,SZ=splat_size;SZ>=1;SZ=SZ/2,++l)					//and floating-point textures. We'll use whichever we want later.
	  createSplatRGBA(SZ,imgb,l);
	
   glBindTexture(GL_TEXTURE_2D, balltex[1]);					//Create mipmap floating-point textures 
   for(int l=0,SZ=splat_size;SZ>=1;SZ=SZ/2,++l)					//
	  createSplatFloat(SZ,img,l);

   delete[] img;
}


void PointCloudSurfaceRenderer::draw(Grid& grid)
{
	PointCloudSurface& surface = (PointCloudSurface&)grid;

    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glEnable(GL_LIGHTING);  

	switch(draw_mode)
	{
	case DRAW_POINTS:
		drawPoints(surface);
		break;
	case DRAW_GEOM_BALLS:
	case DRAW_SPRITE_BALLS:
		drawBalls(surface,false);
		break;
	case DRAW_SPLATS:
		drawSplats(surface);
		break;
	case DRAW_VOXELS:
		drawCubes(surface);
		break;	
	}		
	
	glutSwapBuffers();
}


void PointCloudSurfaceRenderer::drawBalls(PointCloudSurface& surface,bool use_constant_radius)		//Draw the voxel surface as a set of voxel-sized balls
{
	glShadeModel((draw_mode==DRAW_GEOM_BALLS)?GL_SMOOTH:GL_FLAT);	//In splatting mode, we don't need smooth shading (actually, we don't need shading at all)
		
	if (draw_mode==DRAW_SPRITE_BALLS)								//DRAW_SPRITE_BALLS mode:
	{																// -we don't need lighting, we do it ourselves in the textures
		glDisable(GL_LIGHTING);										// -we need texturing of course
		glEnable(GL_TEXTURE_2D);									// -we don't need blending, since we use the alpha test
		glDisable(GL_BLEND);										// -we need a fragment program
		glEnable(GL_ALPHA_TEST);									//Needed to make transparency work
		glAlphaFunc(GL_GREATER,0.1);								//Normally, 0 should be enough here; yet, it creates some ugly thin-black border effects with mipmapping
		int splat_mode = (use_floating_textures)? 1:0;				//See if we use fixed-point or floating-point balls.
		glBindTexture(GL_TEXTURE_2D, balltex[splat_mode]);			//The two have different textures and shaders
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, getScalingMode());
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, getScalingMode());
		glUseProgram(ball_prog);
		glBegin(GL_POINTS); 										//Render points which are converted in quads by geom. shader
	}	
	else //draw_mode==DRAW_GEOM_BALLS								//DRAW_GEOM_BALLS mode: we'll draw shaded spheres
	{
		glEnable(GL_LIGHTING);		
		if (!use_colormap) glColor3fv(def_color);
	}

    float range   = surface.pointScalars().getMax()-surface.pointScalars().getMin();
	float val_min = surface.pointScalars().getMin();
	float col[3];
	
	for(int i=0,iend=surface.numPoints();i<iend;++i)				//Draw all surface points:
	{
		Point3d s; 
		surface.getPoint(i,s.data);
		float radius = surface.getPointRadius(i);
		
		double r   = ((use_constant_radius)? ball_size:radius)+ball_dt_offset; 	
		float  val = (surface.pointScalars().getC0Scalar(i)-val_min)/range;
		
		if (draw_mode==DRAW_SPRITE_BALLS)							//Call shaders to generate the billboard geometry from the passed point, and rasterize it
		{
			glColor4f(0.0f,val,r*zprScale,(float)use_colormap);		//The B color-component passes the splat radius to the shaders
			glVertex3fv(s.data);									//Scaling affects splat's radius; since geom. shader assumes scaling=1.0f, we prescale radius by zprScaling factor.
		}
		else //DRAW_GEOM_BALLS mode:
		{
		   if (use_colormap)
		   {
			 float2rgb(val,col);									//Set color as function of scalar value
			 glColor3fv(col);
		   } 
		   glPushMatrix();										
		   glTranslatef(s.x,s.y,s.z);
		   if (sphere_faces==2) glutSolidCube(r);					//If sphere_faces==2, we actually draw a cube
		   else glutSolidSphere(r,sphere_faces,sphere_faces);		//Classical sphere drawing
		   glPopMatrix();											//Restore modelview matrix (in case we modified it)
		}
	}
	
	if (draw_mode==DRAW_SPRITE_BALLS)								//Finalize drawing and revert to standard settings
	{
		 glEnd();
		 glUseProgram(0);
		 glEnable(GL_LIGHTING);
		 glDisable(GL_TEXTURE_2D);
		 glDisable(GL_ALPHA_TEST);
	}
}



void PointCloudSurfaceRenderer::drawSplats(PointCloudSurface& surface)
{
  glDisable(GL_BLEND);

  float col[4];
  float range   = surface.pointScalars().getMax()-surface.pointScalars().getMin();
  float val_min = surface.pointScalars().getMin();
    
  float psz = zprScale*splat_radius+splat_dt_offset;				//Compute effective splat size. Note that we do NOT adapt the
																	//splat size based on the radii of all points. Doing so is possible,
																	//but (a) is slower, and (b) requires a world-space-to-screen-space
																	//transformation for all points.
  if (psz<1) psz=1;													//Don't let splats be smaller than 1 pixel

  glEnable(GL_POINT_SMOOTH);						
  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glPointSize(psz);	 
  if (!use_colormap) glColor3fv(def_color);
  
  float onorm[3];
  float n[3] = {0,0,1};
  double* minv = zprInverseModelviewMatrix;
  onorm[0] = minv[0]*n[0] + minv[4]*n[1] + minv[8]*n[2];
  onorm[1] = minv[1]*n[0] + minv[5]*n[1] + minv[9]*n[2];
  onorm[2] = minv[2]*n[0] + minv[6]*n[1] + minv[10]*n[2];  

  glBegin(GL_POINTS);	
  for(int pid=0,NP=surface.numPoints();pid<NP;++pid)
  {
	Point3d p; 
	surface.getPoint(pid,p.data);
	
	if (use_colormap)
	{
		float val = (surface.pointScalars().getC0Scalar(pid)-val_min)/range;	
												//Scalar value, normalized in [0,1]
		float2rgb(val,col);						//Set color as function of scalar value
		glColor3fv(col);
	}

												//Get cached normal vector of splat
	Point3d N = surface.pointNormals().getC0Vector(pid);
	
	if (N.x==0 && N.y==0 && N.z==0) N = onorm;	//If no normal was computed for this point, use view-based one  
	else if (N.dot(onorm)<0)					//If the caller gave an orientation-hint for the normal,
    {  N[0] *= -1; N[1] *= -1; N[2] *= -1;  }	//use it (swap computed normal along hint's orientation)
	
	glNormal3fv(N.data);
	glVertex3fv(p.data);	
   }

   glEnd();	
}


void PointCloudSurfaceRenderer::drawCubes(PointCloudSurface& surface)
{
  glDisable(GL_BLEND);
	
  if (!use_colormap) glColor3fv(def_color);
	
  float col[4];
  float range   = surface.pointScalars().getMax()-surface.pointScalars().getMin();
  float val_min = surface.pointScalars().getMin();
      
  for(int pid=0,NP=surface.numPoints();pid<NP;++pid)
  {
	Point3d p; 
	surface.getPoint(pid,p.data);
	float radius = surface.getPointRadius(pid);
	
	if (use_colormap)
	{
		float val = (surface.pointScalars().getC0Scalar(pid)-val_min)/range; 	
												//Scalar value, normalized in [0,1]
		float2rgb(val,col);
		glColor3fv(col);
	}
			
	glTranslatef(p.x,p.y,p.z);
	glutSolidCube(radius);
	glTranslatef(-p.x,-p.y,-p.z);
   }
}


void PointCloudSurfaceRenderer::drawPoints(PointCloudSurface& surface)
{
  glEnable(GL_POINT_SMOOTH);						
  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glPointSize(1.0f);
  glDisable(GL_LIGHTING);
	
  if (!use_colormap) glColor3fv(def_color);
	
  float col[4];
  float range   = surface.pointScalars().getMax()-surface.pointScalars().getMin();
  float val_min = surface.pointScalars().getMin();

  glBegin(GL_POINTS);      
  for(int pid=0,NP=surface.numPoints();pid<NP;++pid)
  {
	Point3d p; 
	surface.getPoint(pid,p.data);
	
	if (use_colormap)
	{
		float val = (surface.pointScalars().getC0Scalar(pid)-val_min)/range; 	
												//Scalar value, normalized in [0,1]
		float2rgb(val,col);
		glColor3fv(col);
	}
	
	glVertex3f(p.x,p.y,p.z);
  }
  glEnd();
}






